/************************************************************
Copyright (c) 1993 by Silicon Graphics Computer Systems, Inc.

Permission to use, copy, modify, and distribute this
software and its documentation for any purpose and without
fee is hereby granted, provided that the above copyright
notice appear in all copies and that both that copyright
notice and this permission notice appear in supporting
documentation, and that the name of Silicon Graphics not be 
used in advertising or publicity pertaining to distribution 
of the software without specific prior written permission.
Silicon Graphics makes no representation about the suitability 
of this software for any purpose. It is provided "as is"
without any express or implied warranty.

SILICON GRAPHICS DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS 
SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY 
AND FITNESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SILICON
GRAPHICS BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL 
DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, 
DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE 
OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

********************************************************/

#ifdef HAVE_DIX_CONFIG_H
#include <dix-config.h>
#endif

#ifdef HAVE_XKB_CONFIG_H
#include <xkb-config.h>
#endif

#include <stdio.h>
#include <ctype.h>

#ifdef HAVE_SHA1_IN_LIBMD
# include <sha1.h>
#else
# include <openssl/sha.h>
#endif

#define	NEED_EVENTS 1
#include <X11/X.h>
#include <X11/Xos.h>
#include <X11/Xproto.h>
#include <X11/keysym.h>
#include <X11/extensions/XKM.h>
#include "inputstr.h"
#include "scrnintstr.h"
#include "windowstr.h"
#define	XKBSRV_NEED_FILE_FUNCS
#include <xkbsrv.h>
#include <X11/extensions/XI.h>
#include "xkb.h"

#include <stdarg.h>
#include <stdio.h>

#if defined(CSRG_BASED) || defined(linux) || defined(__GNU__)
#include <paths.h>
#endif

#define	PRE_ERROR_MSG "\"The XKEYBOARD keymap compiler (xkbcomp) reports:\""
#define	ERROR_PREFIX	"\"> \""
#define	POST_ERROR_MSG1 "\"Errors from xkbcomp are not fatal to the X server\""
#define	POST_ERROR_MSG2 "\"End of messages from xkbcomp\""

static struct {
    SHA_CTX	 sha;
    char	*buffer;
    int		 offset;
    int		 space;
} cookie;

#ifdef WIN32
#include <X11/Xwindows.h>
int 
Win32System(const char *cmdline)
{
    STARTUPINFO si;
    PROCESS_INFORMATION pi;
    DWORD dwExitCode;
    char *cmd = xstrdup(cmdline);

    ZeroMemory( &si, sizeof(si) );
    si.cb = sizeof(si);
    ZeroMemory( &pi, sizeof(pi) );

    if (!CreateProcess(NULL, cmd, NULL, NULL, FALSE, 0, NULL, NULL, &si, &pi)) 
    {
	LPVOID buffer;
	if (!FormatMessage( 
		    FORMAT_MESSAGE_ALLOCATE_BUFFER | 
		    FORMAT_MESSAGE_FROM_SYSTEM | 
		    FORMAT_MESSAGE_IGNORE_INSERTS,
		    NULL,
		    GetLastError(),
		    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		    (LPTSTR) &buffer,
		    0,
		    NULL ))
	{
	    ErrorF("[xkb] Starting '%s' failed!\n", cmdline); 
	}
	else
	{
	    ErrorF("[xkb] Starting '%s' failed: %s", cmdline, (char *)buffer); 
	    LocalFree(buffer);
	}

	xfree(cmd);
	return -1;
    }
    /* Wait until child process exits. */
    WaitForSingleObject( pi.hProcess, INFINITE );

    GetExitCodeProcess( pi.hProcess, &dwExitCode);
    
    /* Close process and thread handles. */
    CloseHandle( pi.hProcess );
    CloseHandle( pi.hThread );
    xfree(cmd);

    return dwExitCode;
}
#undef System
#define System(x) Win32System(x)
#endif

int
Xkb_printf(void *context, const char *format, ...)
{
    va_list	ap;
    int		length = -1;

    if (context == &cookie) {
	for (;;) {
	    va_start(ap, format);
	    length = vsnprintf(cookie.buffer + cookie.offset,
			       cookie.space - cookie.offset,
			       format, ap);
	    if (cookie.offset + length >= cookie.space)
		cookie.buffer = xnfrealloc(cookie.buffer, cookie.space += 4096);
	    else {
		SHA1_Update(&cookie.sha, cookie.buffer + cookie.offset, length);
		cookie.offset += length;
		break;
	    }
	}
	va_end(ap);
    }

    return (length);
}

void *
Xkb_xkbcomp_init(void)
{
    SHA1_Init(&cookie.sha);
    if (cookie.buffer == NULL)
	cookie.buffer = xnfalloc(cookie.space = 4096);
    cookie.offset = 0;

    return (&cookie);
}

void
Xkb_xkbcomp_final(char name[PATH_MAX], unsigned char sha1[42], char *ext)
{
    int			 i;
    static char		 hex[] = "0123456789abcdef";

    SHA1_Final(sha1, &cookie.sha);

    for (i = 19; i >= 0; i--) {
	sha1[i * 2 + 1] = hex[(sha1[i] >> 4) & 0xf];
	sha1[i * 2]     = hex[sha1[i] & 0xf];
    }
    sha1[40] = '\0';

    snprintf(name, PATH_MAX, "%s%s%s", XkmOutputDirectory, sha1, ext);
}

static Bool
XkbDDXCompileKeymapByNames(XkbDescPtr xkb, XkbComponentNamesPtr names,
			   unsigned want, unsigned need,
			   unsigned char *name, int length)
{
    FILE	*fp;
    struct stat	 st;
    Bool	 result;
    char	*cmd, *tmpfile;
    char	 xkmfile[PATH_MAX];

    if (name == NULL || length < 42)
	return (False);

    XkbWriteKeymapForNames(Xkb_xkbcomp_init(), names, xkb, want, need);
    Xkb_xkbcomp_final(xkmfile, name, ".xkm");
    if ((result = access(xkmfile, R_OK) == 0)) {
	if (stat(xkmfile, &st) != 0 || !S_ISREG(st.st_mode))
	    result = False;
    }
    else {
#ifdef WIN32
	tmpfile = Xprintf("%s\\xkb_XXXXXX.tmp", XkbOutputDirectory);
	(void)mktemp(tmpfile);
	if ((fp = fopen(tmpfile, "wb")) != NULL) {
	    fwrite(cookie.buffer, cookie.offset, 1, fp);
	    fclose(fp);
	}
#else
	tmpfile = "-";
#endif
	cmd = Xprintf("\"%sxkbcomp\" "
		      "-w %d "
		      "\"-R%s\" -xkm \"%s\" "
		      "-em1 %s -emp %s -eml %s "
		      "\"%s\"",
		      XkbBinDirectory,
		      xkbDebugFlags < 2 ? 1 : xkbDebugFlags > 10 ? 10 : xkbDebugFlags,
		      XkbBaseDirectory, tmpfile,
		      PRE_ERROR_MSG, ERROR_PREFIX, POST_ERROR_MSG1,
		      xkmfile);
#ifdef WIN32
	if (System(cmd) >= 0)
	    result = True;
	unlink(tmpfile);
#else
	if ((fp = Popen(cmd, "w")) != NULL) {
	    int	offset, bytes;

	    for (offset = 0; offset < cookie.offset; ) {
		bytes = fwrite(cookie.buffer + offset, 1,
			       cookie.offset - offset, fp);
		if (bytes <= 0)
		    break;
		offset += bytes;
	    }
	    result = Pclose(fp) == 0;
	}
#endif
	xfree(cmd);
    }

    if (result == False)
	name[0] = '\0';

    return (result);
}

static FILE *
XkbDDXOpenConfigFile(char *mapName,char *fileNameRtrn,int fileNameRtrnLen)
{
    char	*buf;
    FILE	*file;

    file = NULL;
    if (mapName != NULL) {
	buf = Xprintf("%s%s.xkm", XkmOutputDirectory, mapName);
	if (buf != NULL) {
	    file = fopen(buf, "rb");
	    if (fileNameRtrn != NULL && fileNameRtrnLen > 0) {
		strncpy(fileNameRtrn, buf, fileNameRtrnLen);
		fileNameRtrn[fileNameRtrnLen - 1] = '\0';
	    }
	    xfree(buf);
	}
    }

    return (file);
}

unsigned
XkbDDXLoadKeymapByNames(	DeviceIntPtr		keybd,
				XkbComponentNamesPtr	names,
				unsigned		want,
				unsigned		need,
				XkbDescPtr *		xkbRtrn,
				char *			nameRtrn,
				int 			nameRtrnLen)
{
XkbDescPtr      xkb;
FILE	*	file;
char		fileName[PATH_MAX];
unsigned	missing;

    *xkbRtrn = NULL;
    if ((keybd==NULL)||(keybd->key==NULL)||(keybd->key->xkbInfo==NULL))
	 xkb= NULL;
    else xkb= keybd->key->xkbInfo->desc;
    if ((names->keycodes==NULL)&&(names->types==NULL)&&
	(names->compat==NULL)&&(names->symbols==NULL)&&
	(names->geometry==NULL)) {
        LogMessage(X_ERROR, "XKB: No components provided for device %s\n",
                   keybd->name ? keybd->name : "(unnamed keyboard)");
        return 0;
    }
    else if (!XkbDDXCompileKeymapByNames(xkb,names,want,need,
                                         nameRtrn,nameRtrnLen)){
	LogMessage(X_ERROR, "XKB: Couldn't compile keymap\n");
	return 0;
    }
    file= XkbDDXOpenConfigFile(nameRtrn,fileName,PATH_MAX);
    if (file==NULL) {
	LogMessage(X_ERROR, "Couldn't open compiled keymap file %s\n",fileName);
	return 0;
    }
    missing= XkmReadFile(file,need,want,xkbRtrn);
    if (*xkbRtrn==NULL) {
	LogMessage(X_ERROR, "Error loading keymap %s\n",fileName);
	fclose(file);
	(void) unlink (fileName);
	return 0;
    }
    else {
	DebugF("Loaded XKB keymap %s, defined=0x%x\n",fileName,(*xkbRtrn)->defined);
    }
    fclose(file);
    return (need|want)&(~missing);
}

Bool
XkbDDXNamesFromRules(	DeviceIntPtr		keybd,
			char *			rules_name,
			XkbRF_VarDefsPtr	defs,
			XkbComponentNamesPtr	names)
{
char 		buf[PATH_MAX];
FILE *		file;
Bool		complete;
XkbRF_RulesPtr	rules;

    if (!rules_name)
	return False;

    if (strlen(XkbBaseDirectory) + strlen(rules_name) + 8 > PATH_MAX) {
        LogMessage(X_ERROR, "XKB: Rules name is too long\n");
        return False;
    }
    sprintf(buf,"%srules/%s", XkbBaseDirectory, rules_name);

    file = fopen(buf, "r");
    if (!file) {
        LogMessage(X_ERROR, "XKB: Couldn't open rules file %s\n", buf);
	return False;
    }

    rules = XkbRF_Create(0, 0);
    if (!rules) {
        LogMessage(X_ERROR, "XKB: Couldn't create rules struct\n");
	fclose(file);
	return False;
    }

    if (!XkbRF_LoadRules(file, rules)) {
        LogMessage(X_ERROR, "XKB: Couldn't parse rules file %s\n", rules_name);
	fclose(file);
	XkbRF_Free(rules,True);
	return False;
    }

    memset(names, 0, sizeof(*names));
    complete = XkbRF_GetComponents(rules,defs,names);
    fclose(file);
    XkbRF_Free(rules, True);

    if (!complete)
        LogMessage(X_ERROR, "XKB: Rules returned no components\n");

    return complete;
}
